@@ -2,7 +2,23 @@ from collections import namedtuple
 
 SchemaVersion = namedtuple("SchemaVersion", "subject schema_id schema version")
 
-VALID_LEVELS = ("NONE", "FULL", "FORWARD", "BACKWARD")
+BACKWARD = "BACKWARD"
+BACKWARD_TRANSITIVE = "BACKWARD_TRANSITIVE"
+FORWARD = "FORWARD"
+FORWARD_TRANSITIVE = "FORWARD_TRANSITIVE"
+FULL = "FULL"
+FULL_TRANSITIVE = "FULL_TRANSITIVE"
+NONE = "NONE"
+
+VALID_LEVELS = (
+    BACKWARD,
+    BACKWARD_TRANSITIVE,
+    FORWARD,
+    FORWARD_TRANSITIVE,
+    FULL,
+    FULL_TRANSITIVE,
+    NONE,
+)
 VALID_METHODS = ("GET", "POST", "PUT", "DELETE")
 VALID_AUTH_PROVIDERS = ("URL", "USER_INFO", "SASL_INHERIT")
 
